#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "database/table.h"
#include "database/keytable.h"

#include "grr.h"
#include "l2trigtab.h"
#include "slp.h"

static int nTabEntries = 0;
static L2TENTRY *ttab = 0;
static int TRIGserver = -1;
static int TRIGtable = -1;
static int offline_trig_minor = -100;
static int offline_trig_major = -100;

static int nSymEntries = 0;
static L2SYMBOL *stab = 0;
static int SYMBserver = -1;
static int SYMBtable = -1;
static int offline_symb_minor = -100;
static int offline_symb_major = -100;

#define TT_TABLE_NAME "L2TRIG_0"
#define TT_INITIAL_SIZE 20
#define TT_N_ITEMS 20 /* number of fields in L2TENTRY */

#define TS_TABLE_NAME "L2SYMBOL_0"
#define TS_INITIAL_SIZE 20
#define TS_N_ITEMS 5

#define L2T_FILE_READ_OK -1

/***  trigger table routines  ***/

/***
   *
   * TT_ReadEntry
   *
   * read one row of the trigger table. This is the messy part. If the number of fields
   * changes, TT_N_ITEMS must be reset, the fscanf call must be changed, as well as the 
   * section which assigns the structure elements.
   *
   ***/

static int TT_ReadEntry(FILE *fd,  L2TENTRY *t, int itemNum)
{
  int rc;

  char name[100];
  unsigned int fcs_type;
  unsigned int flt_type;
  unsigned int slt_type;
  unsigned int control;
  unsigned int prescale;
  unsigned int n_ecal;
  unsigned int n_gamma;
  unsigned int n_muon;
  unsigned int n_hipt;
  unsigned int n_ecal_h;
  unsigned int n_gamma_h;
  unsigned int n_muon_h;
  unsigned int n_hipt_h;
  float pt;
  float pt_hi;
  float min_mass;
  float ip_lo;
  float ip_hi;
  float dz_vtx;

  rc = fscanf(fd, "%s %d %d %d %d %d %d %d %d %d %d %d %d %d %f %f %f %f %f %f\n",
  name,
  &fcs_type,
  &flt_type,
  &slt_type,
  &control,
  &prescale,
  &n_ecal,
  &n_gamma,
  &n_muon,
  &n_hipt,
  &n_ecal_h,
  &n_gamma_h,
  &n_muon_h,
  &n_hipt_h,
  &pt,
  &pt_hi,
  &min_mass,
  &ip_lo,
  &ip_hi,
  &dz_vtx
	 );

  if(rc != TT_N_ITEMS) {
    int err;
    if( (err = ferror(fd)) != 0) {
      grr_error(GRR_SFATAL, "L2TT_ReadFromFile: error on read", err);
      if(err == L2T_FILE_READ_OK) err = 100;
      return err;
    } else if(feof(fd))
      return L2T_FILE_READ_OK;

    else {
      char mess[100];
      sprintf(mess,"L2TT_ReadFromFile: error converting item %d of line %d", rc+1, itemNum+1);
      grr_error(GRR_SFATAL, mess, GRR_FISHY);
      return err;
    }
  }

  if(strlen(name) >= L2T_NAME_LENGTH) {
    char mess[200];
    sprintf(mess, "L2TT_ReadFromFile: name %s is too long, truncating", name);
    grr_error(GRR_SINFO, mess, GRR_FISHY);
    name[L2T_NAME_LENGTH-1] = '\0';
  }
  
  
  strcpy(t->name, name);
  t->fcs_type = fcs_type;
  t->flt_type = flt_type;
  t->slt_type = slt_type;
  t->control = control;
  t->prescale = prescale;
  t->n_ecal = n_ecal;
  t->n_gamma = n_gamma;
  t->n_muon = n_muon;
  t->n_hipt = n_hipt;
  t->n_ecal_h = n_ecal_h;
  t->n_gamma_h = n_gamma_h;
  t->n_muon_h = n_muon_h;
  t->n_hipt_h = n_hipt_h;
  t->pt = pt;
  t->pt_hi = pt_hi;
  t->min_mass = min_mass;
  t->ip_lo = ip_lo;
  t->ip_hi = ip_hi;
  t->dz_vtx = dz_vtx;

  return 0;
}

/***
   *
   * TT_WriteEntry
   *
   * Another messy one. Write one row of the trigger table to fd.
   *
   ***/

static int TT_WriteEntry(FILE *fd,  L2TENTRY *t)
{
  int rc;

  rc = fprintf(fd, "%-20s %2d %3d %3d %9d  %4d  %1d  %1d  %1d  %1d    %1d    %1d    %1d    %1d %4.2f   %4.2f  %3.1f %6.4f %6.4f   %4.2f\n",
	       t->name,
	       t->fcs_type,
	       t->flt_type,
	       t->slt_type,
	       t->control,
	       t->prescale,
	       t->n_ecal,
	       t->n_gamma,
	       t->n_muon,
	       t->n_hipt,
	       t->n_ecal_h,
	       t->n_gamma_h,
	       t->n_muon_h,
	       t->n_hipt_h,
	       t->pt,
	       t->pt_hi,
	       t->min_mass,
	       t->ip_lo,
	       t->ip_hi,
	       t->dz_vtx
	       );

  if(rc < 0)
    return GRR_FISHY;

  return GRR_OK;
}

/***
   *
   * L2TT_ReadFromFile
   *
   ***/

int L2TT_ReadFromFile(char *fName, L2TENTRY **tab, int *nEntries)
{
  FILE *fd;
  int rc;
  int nAllocated = 0;

  *nEntries = 0;

  if( (fd = fopen(fName, "r")) == 0) {
    char mess[200] = {"\n"};
    if(fName != 0)
      sprintf(mess,"L2TT_ReadFromFile: unable to open file: %s", fName);

    grr_error(GRR_SFATAL, mess, GRR_FISHY);
    return GRR_FISHY;
  }

  /* initial allocation */
  if( (*tab = (L2TENTRY*) malloc(TT_INITIAL_SIZE * sizeof(L2TENTRY) )) == 0) {
    grr_error(GRR_SFATAL, "L2TT_ReadFromFile: failed in initial memory allocation", GRR_FISHY);
    fclose(fd);
    return GRR_FISHY;
  }
  nAllocated = TT_INITIAL_SIZE;

  {
    int c;
    while( (c=fgetc(fd)) != '\n') ; /* skip column headers */
  }

  while( (rc = TT_ReadEntry( fd, &(*tab)[*nEntries] , *nEntries)) == 0) {
    (*nEntries)++;
    if( *nEntries == nAllocated ) {  /* need some more space */
      nAllocated += TT_INITIAL_SIZE;
      if( (*tab = (L2TENTRY*) realloc(*tab, nAllocated * sizeof(L2TENTRY) )) == 0) {
	grr_error(GRR_SFATAL, "L2TT_ReadFromFile: failed in memory reallocation", GRR_FISHY);
	fclose(fd);
	return GRR_FISHY;
      }
    }
  }

  fclose(fd);

  if(rc == L2T_FILE_READ_OK) {
    if( (*tab = (L2TENTRY*) realloc(*tab, *nEntries * sizeof(L2TENTRY) )) == 0) {
      grr_error(GRR_SFATAL, "L2TT_ReadFromFile: failed while squeezing out unused memory", GRR_FISHY);
      return GRR_FISHY;
    }
    return GRR_OK;
  } else {
    free(*tab);
    return rc;
  }
}


/***
   *
   * L2TT_ReadFromDB
   
   ***/

int L2TT_ReadFromDB(char *dbName, int key, L2TENTRY **tab, int *nEntries, int major, int minor)
{
  int t_id, rc;
  Tables x;
  char mess[100];


  x = OpenTables(dbName, 0);


  if( x == 0) {
    char mess[100];
    sprintf(mess,"L2TT_ReadFromDB: failed to open data base: %s", dbName);
    grr_error(GRR_SFATAL, mess, GRR_FISHY);
    return GRR_FISHY;
  }


  /* the first scheme is to use another number in the table name */
  t_id = BookTable(x, TT_TABLE_NAME, 0, 0);


  if ( t_id == 0) {
    grr_error(GRR_SFATAL, "L2TT_ReadFromDB: error return from BookTable", GRR_FISHY);
    CloseTables(x);
    return GRR_FISHY;
  }


   if(offline_trig_major == -100 && offline_trig_minor == -100) {
      offline_trig_major = major;
      offline_trig_minor = minor;
   }


   if( (offline_trig_major != -1) || (offline_trig_minor != -1) ){
     TableStamp(x, offline_trig_major, offline_trig_minor); 
   }
   else {
     TRIGserver = KeyFindServer(dbName);
     TRIGtable  = KeyFindTable(TRIGserver, TT_TABLE_NAME);
     if (!KeyTableNew(TRIGtable)) {
       CloseTables(x);
       return GRR_OK;
     }

     KeyTableStamp(x, dbName, TT_TABLE_NAME); 
   }

  /* fill and store tables */

  rc = FetchTable(x, t_id, TABLEeq); /* with keytables change to TABLEeq */

  if(rc != t_id) {
    grr_error(GRR_SFATAL, "L2TT_ReadFromDB: error return from FetchTable", GRR_FISHY);
    CloseTables(x);
    return GRR_FISHY;
  }

  {
    int npar, i;
    char **name;
    
    /* go to last field to find number of parameters  (MM: don't know why the skip is needed) */
    SkipTableField(x, TT_N_ITEMS);
    npar = TableFieldSize(x);

    *nEntries = npar;

    name = (char **) malloc(npar * sizeof (char *)); /* need some scratch space to read in names */
    name[0] = 0; /* required due to internal use of realloc */

    if( (*tab = malloc(npar * sizeof(L2TENTRY))) == 0) {
      grr_error(GRR_SFATAL, "L2TT_ReadFromDB: error on memory allocation for trigger table", GRR_FISHY);
      CloseTables(x);
      free(name);
      return GRR_FISHY;
    }

    /* read the data */
    SkipTableField(x,1);

    /* first get the trigger names and copy them into the table */

    if( (rc = GetTableField(x, TABLEstring, npar, name)) != npar) {
      grr_error(GRR_SFATAL, "L2TT_ReadFromDB: error return from GetTableField when reading names", GRR_FISHY);
      CloseTables(x);
      free(name);
      return GRR_FISHY;
    }

    for(i = 0; i < npar; i++) {

      if(name[i] == 0) {
	grr_error(GRR_SWARNING, "L2TT_ReadFromDB: problem with parameter name", GRR_FISHY);
	(*tab)[i].name[0]  = '\0';
	continue;
      }

      strncpy((*tab)[i].name, name[i], L2T_NAME_LENGTH);
    }
    free(name[0]); /* I get a segmentation violation if I free each element */
    free(name);


    /* Now for the rest of the fields. Wouldn't it be nice if we could just get everything with
       one call rather than going through all this nonsense? Should be possible */

#define GetNCheck(type, data, v_name) \
    {int pnc_rc = GetTableStruct(x, (type), sizeof(L2TENTRY), *nEntries, (data)); \
    if(pnc_rc != *nEntries) {  char mess[100]; \
    sprintf(mess,"L2TT_ReadFromDB: GetTableStruct failed for field: %s", (v_name)); \
    CloseTables(x); \
    grr_error(GRR_SFATAL, mess, GRR_FISHY); \
    return GRR_FISHY; }}

    GetNCheck(TABLEubyte,  &(*tab)->fcs_type, "fcs_type");
    GetNCheck(TABLEubyte,  &(*tab)->flt_type, "flt_type");
    GetNCheck(TABLEushort, &(*tab)->slt_type, "slt_type");
    GetNCheck(TABLEuint,   &(*tab)->control, "control");
    GetNCheck(TABLEuint,   &(*tab)->prescale, "prescale");
    GetNCheck(TABLEubyte,  &(*tab)->n_ecal, "n_ecal");
    GetNCheck(TABLEubyte,  &(*tab)->n_gamma, "n_gamma");
    GetNCheck(TABLEubyte,  &(*tab)->n_muon, "n_muon");
    GetNCheck(TABLEubyte,  &(*tab)->n_hipt, "n_hipt");
    GetNCheck(TABLEubyte,  &(*tab)->n_ecal_h, "n_ecal_h");
    GetNCheck(TABLEubyte,  &(*tab)->n_gamma_h, "n_gamma_h");
    GetNCheck(TABLEubyte,  &(*tab)->n_muon_h, "n_muon_h");
    GetNCheck(TABLEubyte,  &(*tab)->n_hipt_h, "n_hipt_h");
    GetNCheck(TABLEfloat,  &(*tab)->pt, "pt");
    GetNCheck(TABLEfloat,  &(*tab)->pt_hi, "pt_hi");
    GetNCheck(TABLEfloat,  &(*tab)->min_mass, "min_mass");
    GetNCheck(TABLEfloat,  &(*tab)->ip_lo, "ip_lo");
    GetNCheck(TABLEfloat,  &(*tab)->ip_hi, "ip_hi");
    GetNCheck(TABLEfloat,  &(*tab)->dz_vtx, "dz_vtx");
    
#undef GetNCheck

  }

  CloseTables(x);  

  return GRR_OK;
}

/***
   *
   * L2TT_WriteToFile
   *
   * If fName is null, write to standard out.
   *
   ***/

int L2TT_WriteToFile(char *fName, L2TENTRY *tab, int nEntries)
{
  FILE *fd;
  int rc, i;

  if(fName != 0) {
    if( (fd = fopen(fName, "w")) == 0) {
      char mess[200] = {"\n"};
      if(fName != 0)
	sprintf(mess,"L2TT_ReadFromFile: unable to open file: %s", fName);

      grr_error(GRR_SFATAL, mess, GRR_FISHY);
      return GRR_FISHY;
    }
  } else

    fd = stdout;

  fprintf(fd,"name                fcs flt slt   control presc ec ga mu hp ec_h ga_h mu_h hp_h   pt  pt_hi mass  ip_lo  ip_hi dz_vtx\n");

  for(i=0; i < nEntries; i++) {
    rc = TT_WriteEntry(fd, tab+i);
    if(rc != GRR_OK) {
      char mess[100];
      sprintf(mess,"L2TT_WriteToFile: error while writing row %d", i);
      grr_error(GRR_SWARNING, mess, GRR_FISHY);
      if(fName != 0) 
	fclose(fd);
      return rc;
    }
  }
  if(fName != 0)
    fclose(fd);

  return GRR_OK;
}

/***
   *
   * L2TT_WriteToDB
   *
   ***/

int L2TT_WriteToDB(char *dbName, int key, L2TENTRY *tab, int nEntries)
{

  int t_id, rc;
  Tables x;

  x = OpenTables(dbName, 0);
  
  if( x == 0) {
    char mess[100];
    sprintf(mess,"L2TT_WriteToDB: failed to open data base: %s", dbName);
    grr_error(GRR_SFATAL, mess, GRR_FISHY);
    return GRR_FISHY;
  }

  /* the first scheme is to use another number in the table name */
  /* If one is required to add a parameter one changes this number */
  t_id = BookTable(x,TT_TABLE_NAME,
	      "A:name UB:fcs UB:flt US:slt UI:control UI:presc UB:ec UB:ga UB:mu UB:hp UB:ecH UB:gaH "
	      "UB:muH UB:hpH F:pt F:ptHi F:mass F:ipLo F:ipHi F:dzVtx",
	      "trigger definition");

  if ( t_id == 0) {
    grr_error(GRR_SFATAL, "L2TT_WriteToDB: error return from BookTable", GRR_FISHY);
    CloseTables(x);
    return GRR_FISHY;
  }

  /* use the major as time and the minor as 0 for online */
  TableStamp(x, -1, 0);

  /* fill and store tables */

  SkipTableField(x, 1);

  {
    /* character strings seem to require special treatment. The following seems to work... */

    char **name;
    int i;
    if( (name = (char**) malloc(nEntries * sizeof(char*))) == 0) {
      grr_error(GRR_SFATAL, "L2TT_WriteToDB: memory allocation for name fields failed", GRR_FISHY);
      CloseTables(x);
      return GRR_FISHY;
    }

    for(i=0; i < nEntries; i++)
      name[i] = tab[i].name;

    if( (rc = PutTableField(x, TABLEstring, nEntries, name)) == 0) {
      grr_error(GRR_SFATAL, "L2TT_WriteToDB: PutTableField failed on name fields", GRR_FISHY);
      CloseTables(x);
      return GRR_FISHY;
    }
    free(name);
  }

#define TT_PutNCheck(type, data, v_name) \
  {int pnc_rc = PutTableStruct(x, (type), sizeof(L2TENTRY), nEntries, (data)); \
  if(pnc_rc != nEntries) {  char mess[100]; \
  sprintf(mess,"L2TT_WriteToDB: PutTableStruct failed for field: %s", (v_name)); \
  CloseTables(x); \
  grr_error(GRR_SFATAL, mess, GRR_FISHY); \
  return GRR_FISHY; }}
  
  /*  TT_PutNCheck(TABLEstring,  ttab->name, "name"); doesn't work ---- see above */
  TT_PutNCheck(TABLEubyte,  &tab->fcs_type, "fcs_type");
  TT_PutNCheck(TABLEubyte,  &tab->flt_type, "flt_type");
  TT_PutNCheck(TABLEushort, &tab->slt_type, "slt_type");
  TT_PutNCheck(TABLEuint,   &tab->control, "control");
  TT_PutNCheck(TABLEuint,   &tab->prescale, "prescale");
  TT_PutNCheck(TABLEubyte,  &tab->n_ecal, "n_ecal");
  TT_PutNCheck(TABLEubyte,  &tab->n_gamma, "n_gamma");
  TT_PutNCheck(TABLEubyte,  &tab->n_muon, "n_muon");
  TT_PutNCheck(TABLEubyte,  &tab->n_hipt, "n_hipt");
  TT_PutNCheck(TABLEubyte,  &tab->n_ecal_h, "n_ecal_h");
  TT_PutNCheck(TABLEubyte,  &tab->n_gamma_h, "n_gamma_h");
  TT_PutNCheck(TABLEubyte,  &tab->n_muon_h, "n_muon_h");
  TT_PutNCheck(TABLEubyte,  &tab->n_hipt_h, "n_hipt_h");
  TT_PutNCheck(TABLEfloat,  &tab->pt, "pt");
  TT_PutNCheck(TABLEfloat,  &tab->pt_hi, "pt_hi");
  TT_PutNCheck(TABLEfloat,  &tab->min_mass, "min_mass");
  TT_PutNCheck(TABLEfloat,  &tab->ip_lo, "ip_lo");
  TT_PutNCheck(TABLEfloat,  &tab->ip_hi, "ip_hi");
  TT_PutNCheck(TABLEfloat,  &tab->dz_vtx, "dz_vtx");

#undef TT_PutNCheck

  if( (rc = StoreTable(x,t_id,0)) != t_id) {
    grr_error(GRR_SWARNING, "L2TT_WriteToDB: error return from StoreTable", GRR_FISHY);
  }

  CloseTables(x);  

  return GRR_OK;
}

/***
   *
   * L2TT_UpdateTable
   *
   ***/

int L2TT_UpdateTable(L2TENTRY *newTab, int nEntries)
{
  if(ttab != 0)
    free(ttab);

  nTabEntries = nEntries;
  ttab = newTab;


  printf("Final de L2TT_UpdateTable: nentries %d \n", nEntries);
  return GRR_OK;
}

/***
   *
   * L2TT_CheckTable
   *
   ***/


int L2TT_CheckTable(L2TENTRY *tab)
{
  return GRR_OK;
}

/***
   *
   * L2TT_CheckEntry
   *
   ***/

int L2TT_CheckEntry(L2TENTRY *entry)
{
  return GRR_OK;
}

/***
   *
   * L2TT_getTrigTable
   *
   ***/


const L2TENTRY *L2TT_getTrigTable(int *nEntries)
{
  *nEntries = nTabEntries;
  return ttab;
}

int L2TT_getTrigEntry(char *name, L2TENTRY **settings)
{
  int i;
  int flag;

  *settings = (L2TENTRY*) malloc(sizeof(L2TENTRY));

  flag = 1;
  for(i = 0; i < nTabEntries; i++)
  {
    if(((!(strcmp(name,(ttab+i)->name))) && ((ttab+i)->control&L2CTL_ENABLE))) 
    {
      *settings = (ttab+i);
      flag = 0;
    }
  }

  return flag;
}




/*** symbol table routines  ***/

static int TS_ReadEntry(FILE *fd,  L2SYMBOL *s, int itemNum)
{
  int rc;

  char group[100];
  char name[100];
  int data_type;

  rc = fscanf(fd,"%s %s %d", group, name, &data_type);

  if(rc != 3) {
    int err;
    if( (err = ferror(fd)) != 0) {
      grr_error(GRR_SFATAL, "L2TS_ReadFromFile: error on read", err);
      if(err == L2T_FILE_READ_OK) err = 100;
      return err;
    } else if(feof(fd))
      return L2T_FILE_READ_OK;

    else {
      char mess[100];
      sprintf(mess,"L2TS_ReadFromFile: error converting item %d of line %d", rc+1, itemNum+1);
      grr_error(GRR_SFATAL, mess, GRR_FISHY);
      return err;
    }
  }

  if( data_type < 0 || data_type > 1) {
      char mess[100];
      sprintf(mess,"L2TS_ReadFromFile: data_type: %d out of range, row: %d ", data_type, itemNum+1);
      grr_error(GRR_SFATAL, mess, GRR_FISHY);
      return GRR_FISHY;
  }

  
  if(strlen(group) >= L2T_NAME_LENGTH) {
    char mess[200];
    sprintf(mess, "L2TS_ReadFromFile: group name %s is too long, truncating", group);
    grr_error(GRR_SINFO, mess, GRR_FISHY);
    group[L2T_NAME_LENGTH-1] = '\0';
  }
  strcpy(s->group, group);

  if(strlen(name) >= L2T_NAME_LENGTH) {
    char mess[200];
    sprintf(mess, "L2TS_ReadFromFile: parameter name %s of group %s is too long, truncating", name, group);
    grr_error(GRR_SINFO, mess, GRR_FISHY);
    name[L2T_NAME_LENGTH-1] = '\0';
  }
  strcpy(s->name, name);

  s->data_type = data_type;

  switch(data_type) {

  case 0:
    fscanf(fd, "%d\n", &(s->i_value));
    s->f_value = 0.;
    break;

  case 1:
    fscanf(fd, "%f\n", &(s->f_value));
    s->i_value = 0;
    break;

  }
  return GRR_OK;

}

static int TS_WriteEntry(FILE *fd,  L2SYMBOL *s)
{

  fprintf(fd,"%-20s %-20s  %d  ", s->group, s->name, s->data_type);

  switch(s->data_type) {

  case 0:
    fprintf(fd, "%10d\n", s->i_value);
    break;

  case 1:
    fprintf(fd, "          %10.4f\n", s->f_value);

  }

  return GRR_OK;

}
int L2TS_ReadFromFile(char *fName, L2SYMBOL **tab, int *nEntries)
{
  FILE *fd;
  int rc;
  int nAllocated = 0;

  *nEntries = 0;

  if( (fd = fopen(fName, "r")) == 0) {
    char mess[200] = {"\n"};
    if(fName != 0)
      sprintf(mess,"L2TS_ReadFromFile: unable to open file: %s", fName);

    grr_error(GRR_SFATAL, mess, GRR_FISHY);
    return GRR_FISHY;
  }

  /* initial allocation */
  if( (*tab = (L2SYMBOL*) malloc(TS_INITIAL_SIZE * sizeof(L2SYMBOL) )) == 0) {
    grr_error(GRR_SFATAL, "L2TS_ReadFromFile: failed in initial memory allocation", GRR_FISHY);
    fclose(fd);
    return GRR_FISHY;
  }
  nAllocated = TS_INITIAL_SIZE;

  {
    int c;
    while( (c=fgetc(fd)) != '\n') ; /* skip column headers */
  }

  while( (rc = TS_ReadEntry( fd, &(*tab)[*nEntries] , *nEntries)) == 0) {
    (*nEntries)++;
    if( *nEntries == nAllocated ) {  /* need some more space */
      nAllocated += TS_INITIAL_SIZE;
      if( (*tab = (L2SYMBOL*) realloc(*tab, nAllocated * sizeof(L2SYMBOL) )) == 0) {
	grr_error(GRR_SFATAL, "L2TS_ReadFromFile: failed in memory reallocation", GRR_FISHY);
	fclose(fd);
	return GRR_FISHY;
      }
    }
  }

  fclose(fd);

  if(rc == L2T_FILE_READ_OK) {
    if( (*tab = (L2SYMBOL*) realloc(*tab, *nEntries * sizeof(L2SYMBOL) )) == 0) {
      grr_error(GRR_SFATAL, "L2TS_ReadFromFile: failed while squeezing out unused memory", GRR_FISHY);
      return GRR_FISHY;
    }
    return GRR_OK;
  } else {
    free(*tab);
    return rc;
  }
}

int L2TS_ReadFromDB(char *dbName, int key, L2SYMBOL **tab, int *nEntries, int major, int minor)
{
  int t_id, rc;
  Tables x;

  x = OpenTables(dbName, 0);

  if( x == 0) {
    char mess[100];
    sprintf(mess,"L2TS_ReadFromDB: failed to open data base: %s", dbName);
    grr_error(GRR_SFATAL, mess, GRR_FISHY);
    return GRR_FISHY;
  }


  /* the first scheme is to use another number in the table name */
  t_id = BookTable(x, TS_TABLE_NAME, 0, 0);

  if ( t_id == 0) {
    grr_error(GRR_SFATAL, "L2TS_ReadFromDB: error return from BookTable", GRR_FISHY);
    CloseTables(x);
    return GRR_FISHY;
  }


   if(offline_symb_major == -100 && offline_symb_minor == -100) {
      offline_symb_major = major;
      offline_symb_minor = minor;
   }


   if( (offline_symb_major != -1) || (offline_symb_minor != -1) ){
     TableStamp(x, offline_symb_major, offline_symb_minor); 
   }
   else {
     SYMBserver = KeyFindServer(dbName);
     SYMBtable  = KeyFindTable(SYMBserver, TS_TABLE_NAME);
     if (!KeyTableNew(SYMBtable)) {
       CloseTables(x);
       return GRR_OK;
     }

     KeyTableStamp(x, dbName, TS_TABLE_NAME); 
   }

  rc = FetchTable(x, t_id, TABLEeq); /* with keytables change to TABLEeq */

  if(rc != t_id) {
    grr_error(GRR_SFATAL, "L2TS_ReadFromDB: error return from FetchTable", GRR_FISHY);
    CloseTables(x);
    return GRR_FISHY;
  }

  {
    int npar, i;
    char **str;
    
    /* go to last field to find number of parameters  (MM: don't know why the skip to last is needed) */
    SkipTableField(x, TS_N_ITEMS);
    npar = TableFieldSize(x);

    *nEntries = npar;

    str = (char **) malloc(npar * sizeof (char *)); /* need some scratch space to read in names */
    str[0] = 0; /* required due to internal use of realloc */

    if( (*tab = malloc(npar * sizeof(L2SYMBOL))) == 0) {
      grr_error(GRR_SFATAL, "L2TS_ReadFromDB: error on memory allocation for trigger table", GRR_FISHY);
      CloseTables(x);
      free(str);
      return GRR_FISHY;
    }

    /* read the data */
    SkipTableField(x,1);

    /* first get the group and parameter names and copy them into the table */

    /* group names */
    if( (rc = GetTableField(x, TABLEstring, npar, str)) != npar) {
      grr_error(GRR_SFATAL, "L2TS_ReadFromDB: error return from GetTableField when reading group names", GRR_FISHY);
      CloseTables(x);
      free(str);
      return GRR_FISHY;
    }

    for(i = 0; i < npar; i++) {

      if(str[i] == 0) {
	grr_error(GRR_SWARNING, "L2TS_ReadFromDB: problem with group name", GRR_FISHY);
	(*tab)[i].group[0]  = '\0';
	continue;
      }

      strncpy((*tab)[i].group, str[i], L2T_NAME_LENGTH);
    }
    free(str[0]); /* I get a segmentation violation if I free each element */
    str[0] = 0;

    /* parameter names */
    if( (rc = GetTableField(x, TABLEstring, npar, str)) != npar) {
      grr_error(GRR_SFATAL, "L2TS_ReadFromDB: error return from GetTableField when reading parameter names", GRR_FISHY);
      CloseTables(x);
      free(str);
      return GRR_FISHY;
    }

    for(i = 0; i < npar; i++) {

      if(str[i] == 0) {
	grr_error(GRR_SWARNING, "L2TS_ReadFromDB: problem with parameter name", GRR_FISHY);
	continue;
      }

      strncpy((*tab)[i].name, str[i], L2T_NAME_LENGTH);
    }
    free(str[0]); /* I get a segmentation violation if I free each element */
    str[0] = 0;

    free(str);


    /* Now for the rest of the fields. Wouldn't it be nice if we could just get everything with
       one call rather than going through all this nonsense? Should be possible */

#define TS_GetNCheck(type, data, v_name) \
    {int pnc_rc = GetTableStruct(x, (type), sizeof(L2SYMBOL), *nEntries, (data)); \
    if(pnc_rc != *nEntries) {  char mess[100]; \
    sprintf(mess,"L2TS_ReadFromDB: GetTableStruct failed for field: %s", (v_name)); \
    CloseTables(x); \
    grr_error(GRR_SFATAL, mess, GRR_FISHY); \
    return GRR_FISHY; }}

    TS_GetNCheck(TABLEint,    &(*tab)->data_type, "data_type");
    TS_GetNCheck(TABLEint,    &(*tab)->i_value, "i_value");
    TS_GetNCheck(TABLEfloat,  &(*tab)->f_value, "f_value");
    
#undef TS_GetNCheck

  }

  CloseTables(x);  

  return GRR_OK;
}

int L2TS_WriteToFile(char *fName, L2SYMBOL *tab, int nEntries)
{
  FILE *fd;
  int rc, i;

  if(fName != 0) {
    if( (fd = fopen(fName, "w")) == 0) {
      char mess[200] = {"\n"};
      if(fName != 0)
	sprintf(mess,"L2TS_ReadFromFile: unable to open file: %s", fName);

      grr_error(GRR_SFATAL, mess, GRR_FISHY);
      return GRR_FISHY;
    }
  } else
    
    fd = stdout;

  fprintf(fd,"group                name                type    i_value   f_value\n");

  for(i=0; i < nEntries; i++) {
    rc = TS_WriteEntry(fd, tab+i);
    if(rc != GRR_OK) {
      char mess[100];
      sprintf(mess,"L2TS_WriteToFile: error while writing row %d", i);
      grr_error(GRR_SWARNING, mess, GRR_FISHY);
      if(fName != 0)
	fclose(fd);
      return rc;
    }
  }

  if(fName != 0)
    fclose(fd);

  return GRR_OK;

}


int L2TS_WriteToDB(char *dbName, int key, L2SYMBOL *tab, int nEntries)
{

  int t_id, rc;
  Tables x;

  x = OpenTables(dbName, 0);
  
  if( x == 0) {
    char mess[100];
    sprintf(mess,"L2TS_WriteToDB: failed to open data base: %s", dbName);
    grr_error(GRR_SFATAL, mess, GRR_FISHY);
    return GRR_FISHY;
  }

  /* the first scheme is to use another number in the table name */
  /* If one is required to add a parameter one changes this number */
  t_id = BookTable(x, TS_TABLE_NAME,
	      "A:group A:name I:type I:iValue F:fValue",
	      "symbol table");

  if ( t_id == 0) {
    grr_error(GRR_SFATAL, "L2TS_WriteToDB: error return from BookTable", GRR_FISHY);
    CloseTables(x);
    return GRR_FISHY;
  }

  /* use the major as time and the minor as 0 for online */
  TableStamp(x, -1, 0);

  /* fill and store tables */

  SkipTableField(x, 1);

  {
    /* character strings seem to require special treatment. The following seems to work... */

    char **str;
    int i;
    if( (str = (char**) malloc(nEntries * sizeof(char*))) == 0) {
      grr_error(GRR_SFATAL, "L2TS_WriteToDB: memory allocation for string variables failed", GRR_FISHY);
      CloseTables(x);
      return GRR_FISHY;
    }
    /* first the group names */
    for(i=0; i < nEntries; i++)
      str[i] = tab[i].group;

    if( (rc = PutTableField(x, TABLEstring, nEntries, str)) == 0) {
      grr_error(GRR_SFATAL, "L2TS_WriteToDB: PutTableField failed on group fields", GRR_FISHY);
      CloseTables(x);
      return GRR_FISHY;
    }

    /* now the parameter names */
    for(i=0; i < nEntries; i++)
      str[i] = tab[i].name;

    if( (rc = PutTableField(x, TABLEstring, nEntries, str)) == 0) {
      grr_error(GRR_SFATAL, "L2TS_WriteToDB: PutTableField failed on name fields", GRR_FISHY);
      CloseTables(x);
      return GRR_FISHY;
    }

    free(str);
  }

#define TS_PutNCheck(type, data, v_name) \
  {int pnc_rc = PutTableStruct(x, (type), sizeof(L2SYMBOL), nEntries, (data)); \
  if(pnc_rc != nEntries) {  char mess[100]; \
  sprintf(mess,"L2TS_WriteToDB: PutTableStruct failed for field: %s", (v_name)); \
  CloseTables(x); \
  grr_error(GRR_SFATAL, mess, GRR_FISHY); \
  return GRR_FISHY; }}
  
  TS_PutNCheck(TABLEint,   &tab->data_type, "type");
  TS_PutNCheck(TABLEint,   &tab->i_value, "i_value");
  TS_PutNCheck(TABLEfloat,  &tab->f_value, "f_value");

#undef TS_PutNCheck

  if( (rc = StoreTable(x, t_id, 0)) != t_id) {
    grr_error(GRR_SWARNING, "L2TS_WriteToDB: error return from StoreTable", GRR_FISHY);
  }

  CloseTables(x);  

  return GRR_OK;
}

int L2TS_UpdateTable(L2SYMBOL *newTab, int nEntries)
{
  if(stab != 0)
    free(stab);

  printf("L2TS_UpdateTable. nEntries %d \n", nEntries);

  nSymEntries = nEntries;
  stab = newTab;

  return GRR_OK;
}

int L2TS_CheckTable(L2SYMBOL *tab)
{
  return GRR_OK;
}

int L2TS_CheckEntry(L2SYMBOL *entry)
{
  return GRR_OK;
}

const L2SYMBOL *L2TS_getTrigTable(int *nEntries)
{
  *nEntries = nSymEntries;
  return stab;
}

const L2SYMBOL *L2TS_getGroup(char *group, int *nEntries)
{
  return 0;
}

const L2SYMBOL *L2TS_getEntry(char *group, char *name)
{
  return 0;
}

int L2TS_getIntValue(char *group, char *name)
{
  int value;
  int i;
  int flag;
  char mess[200];

  flag = 0;
  for(i = 0; i<nSymEntries; i++){
    if((!(strcmp((stab+i)->group,group))) && (!(strcmp((stab+i)->name,name)))){
      if((stab+i)->data_type == 0){
        value = (stab+i)->i_value;
      }
      else{
        sprintf(mess,"L2TS_getIntValue: error in the type of value \n");
        grr_error(GRR_SFATAL,mess,1);
        exit(EXIT_FAILURE);
      }
      flag = 1;
    }
  }

  if(flag == 0)
  {
    sprintf(mess,"L2TS_getIntValue: Requiring an entry that does not exist \n");
    grr_error(GRR_SFATAL,mess,1);
    exit(EXIT_FAILURE);
  }

  return value;
}

float L2TS_getFloatValue(char *group, char *name)
{
  float value;
  int i;
  int flag;
  char mess[200];

  flag = 0;
  for(i = 0; i<nSymEntries; i++){
    if((!(strcmp((stab+i)->group,group))) && (!(strcmp((stab+i)->name,name)))){
      if((stab+i)->data_type == 1){
        value = (stab+i)->f_value;
      }
      else{
        sprintf(mess,"L2TS_getFloatValue: error in the type of value \n");
        grr_error(GRR_SFATAL,mess,1);
        exit(EXIT_FAILURE);
      }
      flag = 1;
    }
  }
  if(flag == 0)
  {
    sprintf(mess,"L2TS_getFloatValue: Requiring an entry that does not exist \n");
    grr_error(GRR_SFATAL,mess,1);
    exit(EXIT_FAILURE);
  }

  return value;
}

