
/**************************

16.4.98

Need a general way of excluding particular gede (plane) records from the
plane table: a way to specify which planes are to be excluded, a way to
exclude them.

Specification:

void Pl_exclude(char sl_name[4], int sector, int layer)

sl_name must be a valid super layer name

sector must be either a valid sector number for the superlayer in
question or 0 to exclude all sectors.

layer must be either a vaild layer number for the superlayer in
question or 0 to exlcude all layers.

The call:

 Pl_exclude("SI01", 0, 0)

will exclude all of SI01 and has the same effect on the plane table
as excluding the superlayer from the superlayer table. The only
difference is that the superlayer record will appear in the superlayer
table if the above call is used, and will not appear if the superlayer
exclusion function is used. I don't believe this difference will lead
to any significant difference in behavior.


====

How to implement.

Just keep track of all the exclusion statements. When an attempt is
made to insert a plane, see if it's in the set of excluded planes.
If so, don't allow insertion.

Since a large number of records could in principle be in the exclusion
table, I'll be a little less quick than I was with the superlayer
excluder. I'll build it in a separate file but put the interface into
pltable.h.

***********************************/

#include <stdlib.h>
#include <stdio.h>
#include "string.h"
#include "grr.h"
#include "sltable.h"
#include "pltable.h"

typedef struct {
  char sl[4];
  int  sector;
  int  layer;
} EXCL;

static int nElements = 0;
static int nSlots = 0;
static const int allocSlots = 20;

static EXCL *exTab = NULL;

void L2_excludePlane(const char sl_name[4], int sector, int layer)
{
  EXCL *tmp;
  char mess[200];

  /*** Allocate space if needed ***/

  if( nElements >= nSlots ) {
    if(nSlots == 0) { /* (I've had trouble reallocing NULL pointers with some compilers.)*/
      if( (exTab = (EXCL *)malloc( allocSlots * sizeof(EXCL) )) == NULL) {
	sprintf(mess,"l2simu-Pl_exclude: unable to allocate \n");
	grr_error(GRR_SFATAL,mess,1);
	return;
      }
      nSlots = allocSlots;
    } else {
      if( (tmp = (EXCL *) realloc( exTab, (nSlots + allocSlots) * sizeof(EXCL) )) == NULL) {
	sprintf(mess,"l2simu-Pl_exclude: unable to re-allocate \n");
	grr_error(GRR_SFATAL,mess,1);
	return;
      }
      nSlots += allocSlots;
      exTab = tmp;
    }
  }

  /* We could make a lot of checks for whacky calls. Let's call it a
     possible future enhancement.
     */

  strncpy( exTab[nElements].sl, sl_name, 4 );
  exTab[nElements].sector = sector;
  exTab[nElements].layer  = layer;
  nElements++;


}

/***
   *
   * return 0 if not on exclusion list, 1 if on exclusion list.
   *
   ***/

int Pl_isExcluded(const PLANE *pl)
{
  int i;
  SUPERLAYER *sl;
  
  if( pl == NULL)
    return 0;

  sl = SLLookup_pl(pl);

  if(sl == NULL)
    return 0;

  for(i = 0; i < nElements; i++) {
    
    if(strncmp(sl->name, exTab[i].sl, 4) != 0)
      continue;

    if(exTab[i].sector != 0 && exTab[i].sector != pl->sec)
      continue;

    if(exTab[i].layer != 0 && exTab[i].layer != pl->lay)
      continue;

    return 1;
  }

  return 0;
}

void Pl_PrintExcludeTable(void)
{
  int i;

  printf("Table of excluded planes\n");

  for(i = 0; i < nElements; i++) 
    printf("%d, superlayer: %4s, sector: %d, layer: %d\n",
	   i, exTab[i].sl, exTab[i].sector, exTab[i].layer);

}







