#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sltable.h"
#include "grr.h"

typedef SUPERLAYER Item;
typedef struct melement{
	Item *item;
	struct melement *next;
} Element;

static int numElements = 0;
/*static char *listType = "module type";*/
static Element *iteration;

static Element *headElement = NULL;
static Element *lastElement = NULL;

/***
   * Insertions of any superlayers on the "exclude" list are disallowed. This
   * will cause planes in those superlayers to be excluded from the plane geometry 
   * table and data in those planes not to be inserted in the plane-data table.
   *
   ***/

static int nExclusions = 0;
#define SL_MAX_EXCLUSIONS 100
static char exclude[SL_MAX_EXCLUSIONS][5];

void L2_excludeSL( const char sl[5] )
{
  char mess[200];
  if(nExclusions >= SL_MAX_EXCLUSIONS) {
    sprintf(mess,"L2_excludeSL: no more room in exclusion table\n"); 
    grr_error(GRR_SINFO,mess,0);
    return;
  }
  strncpy(exclude[nExclusions++], sl, 5);

  return;
}

Item *SLLookup_pl( const PLANE *pl)
{
  if(pl == NULL)
    return NULL;

  return SLLookup_bp(pl->cmp, pl->l2npl);
}

Item *SLLookup_bp(int cmp, int plane )  /* get superlayer record, given plane number */
{

	Element *anElement = headElement;
	Item *item;
	
	for(anElement = headElement; anElement != NULL; anElement = anElement->next) {
		item = anElement->item;
		if( cmp == item->cmp && (plane >= item->fpl && plane < (item->fpl + item->npl))) {
			return item;
		}
	}
	
	return NULL;
}

/***
 *
 * SLLookup_bn
 *
 * Given the 4 character arte name, return a pointer to the superlayer record.
 *
 ***/

SUPERLAYER *SLLookup_bn(const char name[4])
{
  Element *anElement = headElement;
  Item *item;
	
  for(anElement = headElement; anElement != NULL; anElement = anElement->next) {
    item = anElement->item;
    if( *( (int *)name) == *( (int *)item->name)) {
      return item;
    }
  }
  
  return NULL;
}


/***
 *
 * SL_IterateInit
 *
 * Set iteration to point to the first element of the list. Return it's iterm
 *
 ***/

Item *SL_IterateInit()
{
  iteration = headElement;

  return (iteration != NULL)? iteration->item : NULL;
}

/***
 *
 * SL_IterateNext
 *
 * return next element on list, increment the iteration.
 *
 ***/

Item *SL_IterateNext()
{
  if(iteration == NULL)
    return NULL;

  iteration = iteration->next;

  return (iteration != NULL)? iteration->item : NULL;
}

void SLInsert(Item *theItem)
{
	Element *newElement;

	int i;

	for(i = 0; i < nExclusions; i++ )
	  if(strncmp(exclude[i], theItem->name, 4) == 0) {
	    /*
	    	    printf("l2: superlayer %s excluded from l2 superlayer table\n", exclude[i]);
		    */
	    return;
	  }

	numElements++;
	
	newElement = malloc( sizeof(Element));
	
	if(headElement == NULL) {
		headElement = newElement;
		headElement->next = NULL;
		lastElement = headElement;
	} else {
		lastElement->next = newElement;
		lastElement = lastElement->next;
		lastElement->next = NULL;
	}
		lastElement->item = theItem;
}

int SLGetNbEntries()
{
  Element *anElement;
  int nentries = 0;
  for( anElement = headElement; anElement != NULL; anElement = anElement->next) {
    nentries++;
  }
  return nentries;
}

void SL_destroy(void)
{
  struct melement *elem, *next;

  elem = headElement;

  while( elem != NULL ) {
    free(elem->item);
    next = elem->next;
    free(elem);
    elem = next;
  }

  numElements = 0;
  headElement = NULL;
  lastElement = NULL;
  iteration = NULL;
}

/***
 *
 * Sl_formatCNA
 *
 * Fill the superlayer section of L2FORMMAP 
 * 
 *
 ***/
void SL_formatCNA(SUPERLAYER * splr)
{
        Element *anElement;
        
        for( anElement = headElement; anElement != NULL; anElement = anElement->next) {
	  *splr = *anElement->item;
	  splr++;
        }
}

/***
 *
 * SL_PrintTable
 *
 * Print out the super layer table,
 * e.g. for generation of an ascii file
 *
 ***/

void SL_PrintTable(FILE *fd)
{
  Item *item;
	
  if( fd == NULL) {
    fprintf(fd, "SL_PrintTable: null file descriptor\n");
    exit(1);
  }

  fprintf(fd,"\\SUPERLAYERS %-d\n",  numElements);
  fprintf(fd,"l2nsl name cmp nsl mfl lay  ");
  fprintf(fd,"xo        yo        z1        z2        npl  fpl\n");

  for(item = SL_IterateInit(); item != NULL; item = SL_IterateNext()) {
    SL_PrintSuperLayer(item, fd);
  }
}
/***
 *
 * SL_PrintSuperLayer
 *
 * Print out the geometry of one superlayer
 *
 ***/
void SL_PrintSuperLayer(Item *sl, FILE *fd)
{
  
  if( sl == NULL) {
    fprintf(stderr,"SL_PrintSuperLayer: asked to print NULL superlayer\n");
    exit(1);
  }
  fprintf(fd,"%-5d %c%c%c%c %-3d %-3d %-3d %-3d %-9.4f %-9.4f %-9.4f %-9.4f %-4d %-4d \n",
  sl->l2nsl, sl->name[0], sl->name[1], sl->name[2], sl->name[3], sl->cmp, sl->nsl, sl->mfl, sl->lay, 
  sl->xo,    sl->yo,   sl->z1,  sl->z2, sl->npl, sl->fpl);
}


/***
 *
 * SL_ReadSuperLayer 
 *
 * Read the geometry of one superlayer from ascii file
 *
 ***/
void SL_ReadSuperLayer(FILE *fd, SUPERLAYER **Psl)
{
    int c;
    SUPERLAYER *sl;
    char mess[200];
    
    sl = malloc( sizeof(SUPERLAYER));
    if(sl == NULL){
      sprintf(mess,"sltable: unable to allocate memory for superlayers\n"); 
      grr_error(GRR_SFATAL,mess,1);
      exit(EXIT_FAILURE);
    }

    c = fscanf(fd, "%d %c%c%c%c %d %d %d %d %f %f %f %f  %d %d ", 
    &sl->l2nsl, &sl->name[0], &sl->name[1], &sl->name[2], &sl->name[3], 
	       &sl->cmp, &sl->nsl, &sl->mfl, &sl->lay, 
    &sl->xo, &sl->yo, &sl->z1, &sl->z2, &sl->npl, &sl->fpl);
    
    if( c == EOF) {
      sprintf(mess,"Sltable: error  while reading super layer records\n"); 
      grr_error(GRR_SFATAL,mess,1);
      exit(EXIT_FAILURE);
    }
 
    *Psl = sl;
}

/***
 *
 * SL_PrintTableSpy
 *
 * Print out the super layer table as an initialization statement of
 * array "slayer[]".
 *
 ***/

void SL_PrintTableSpy(FILE *fd)
{
  Item *sl;
	
  if( fd == NULL) {
    fprintf(fd, "SL_PrintTable: null file descriptor\n");
    exit(1);
  }

  fprintf(fd, "#define MAX_SUPER_LAYERS %-d\n\n", numElements);
  fprintf(fd, "static SUPERLAYER slayer[] = {\n");

  for(sl = SL_IterateInit(); sl != NULL; sl = SL_IterateNext()) {
      fprintf(fd,"%2d, %2d, %3d, %3d, %4d, %4d, %9.4f, %9.4f,\n",
	  sl->l2nsl, sl->cmp, sl->nsl, sl->lay, sl->npl, sl->fpl, sl->z1, sl->z2);
  }

  fprintf(fd, "};\n" );
}
